﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.Drawing;


///////////////////////////////////////////////////////////////////////////////
//
//  www.interestprograms.ru /программы, игры и их исходные коды/
//  Протокол TCP серверное приложение. Часть 1.
//
///////////////////////////////////////////////////////////////////////////////

namespace tcpserver
{

    public partial class TcpServerForm : Form
    {
        public TcpServerForm()
        {
            InitializeComponent();

            listBox1.HorizontalScrollbar = true;
        }

        // Высокоуровневая надстройка для прослушивающего сокета
        TcpListener _server;

        // Количество принимаемых подключений к серверу
        const int MAXNUMCLIENTS = 5;

        // Высокоуровневая надстройка для сокетов обмена сообщений с 
        // клиентскими приложениями.
        TcpClient[] clients = new TcpClient[MAXNUMCLIENTS];

        // Счетчик подключенных клиентов
        int _countClient = 0;

        // Флаг мягкой остановки циклов и дополнительных потоков
        bool _stopNetwork;
        


        #region Управление серверным приложением

        private void buttonStart_Click(object sender, EventArgs e)
        {
            StartServer();
        }


        private void buttonSend_Click(object sender, EventArgs e)
        {
            string s = "сервер" + ": " + textBoxSend.Text;
            SendToClients(s, -1);
        }


        private void buttonStop_Click(object sender, EventArgs e)
        {
            StopServer();
        }

        private void TcpServerForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            StopServer();
        }

        #endregion


        #region Функциональная часть сетевой работы


        // Запуск сервера и вспомогательного потока акцептирования клиентских подключений
        // т.е. назначения сокетов ответственных за обмен сообщениями 
        // с соответствующим клиентским приложением
        void StartServer()
        {
            // Предотвратим повторный запуск сервера
            if (_server == null)
            {
                // Блок перехвата исключений на случай запуска одновременно
                // двух серверных приложений с одинаковым портом.
                try
                {
                    _stopNetwork = false;
                    _countClient = 0;
                    UpdateClientsDisplay();

                    int port = int.Parse(textBoxPort.Text);
                    _server = new TcpListener(IPAddress.Any, port);
                    _server.Start();


                    Thread acceptThread = new Thread(AcceptClients);
                    acceptThread.Start();

                    // Визуальное оповещение, что сервер запущен
                    this.BackColor = Color.FromArgb(150, 192, 255);

                }
                catch
                {
                    ErrorSound();
                }
            }
        }


        // Принудительная остановка сервера и запущенных потоков.
        void StopServer()
        {
            if (_server != null)
            {
                _server.Stop();
                _server = null;
                _stopNetwork = true;

                for (int i = 0; i < MAXNUMCLIENTS; i++)
                {
                    if (clients[i] != null) clients[i].Close();
                }

                // Визуально оповещаем, что сервер остановлен.
                this.BackColor = Color.FromName("Control");
            }
        }

        // Принимаем запросы клиентов на подключение и
        // привязываем к каждому подключившемуся клиенту 
        // сокет (в данном случае объект класса TcpClient)
        // для обменом сообщений.
        void AcceptClients()
        {
            while (true)
            {
                try
                {
                    this.clients[_countClient] = _server.AcceptTcpClient();
                    Thread readThread = new Thread(ReceiveRun);
                    readThread.Start(_countClient);
                    _countClient++;


                    // Данный метод, хотя и вызывается в отдельном потоке (не в главном),
                    // но находит родительский поток и выполняет делегат указанный в качестве параметра 
                    // в главном потоке, безопасно обновляя интерфейс формы.
                    Invoke(new UpdateClientsDisplayDelegate(UpdateClientsDisplay));
                }
                catch
                {
                    // Перехватим возможные исключения
                    ErrorSound();
                }


                if (_countClient == MAXNUMCLIENTS || _stopNetwork == true)
                {
                    break;
                }

            }
        }


        /// <summary>
        /// Отправка сообщений клиентам
        /// </summary>
        /// <param name="text">текст сообщения</param>
        /// <param name="skipindex">индекс клиента которому не посылается сообщение</param>
        void SendToClients(string text, int skipindex)
        {
            for (int i = 0; i < MAXNUMCLIENTS; i++)
            {
                if (clients[i] != null)
                {
                    if (i == skipindex) continue;


                    // Подготовка и запуск асинхронной отправки сообщения.
                    NetworkStream ns = clients[i].GetStream();
                    byte[] myReadBuffer = Encoding.Default.GetBytes(text);
                    ns.BeginWrite(myReadBuffer, 0, myReadBuffer.Length,
                                                                 new AsyncCallback(AsyncSendCompleted), ns);
                }
            }
        }

        // Асинхронная отправка сообщения клиенту.
        public void AsyncSendCompleted(IAsyncResult ar)
        {
            NetworkStream ns = (NetworkStream)ar.AsyncState;
            ns.EndWrite(ar);
        }


        // Извлечение сообщения от клиента и ретрансляция полученного 
        // сообщения другим клиентам
        void ReceiveRun(object num)
        {
            while (true)
            {
                try
                {
                    string s = null;
                    NetworkStream ns = clients[(int)num].GetStream();

                    // Раскомментировав строчку ниже, тем самым уменьшив размер приемного буфера, можно убедиться,
                    // что прием данных будет все равно осуществляться полностью.
                    //clients[(int)num].ReceiveBufferSize = 2;
                    while (ns.DataAvailable == true)
                    {
                        // Определить точный размер буфера приема позволяет свойство класса TcpClient - Available
                        byte[] buffer = new byte[clients[(int)num].Available];

                        ns.Read(buffer, 0, buffer.Length);
                        s += Encoding.Default.GetString(buffer);
                    }

                    if (s != null)
                    {
                        // Данный метод, хотя и вызывается в отдельном потоке (не в главном),
                        // но находит родительский поток и выполняет делегат указанный в качестве параметра 
                        // в главном потоке, безопасно обновляя интерфейс формы.
                        Invoke(new UpdateReceiveDisplayDelegate(UpdateReceiveDisplay), new object[] { (int)num, s });

                        // Принятое сообщение от клиента перенаправляем всем клиентам
                        // кроме текущего.
                        s = "№" + ((int)num).ToString() + ": " + s;
                        SendToClients(s, (int)num);
                        s = String.Empty;
                    }

                    // Вынужденная строчка для экономия ресурсов процессора.
                    // Неизящный способ.
                    Thread.Sleep(100);
                }
                catch
                {
                    // Перехватим возможные исключения
                    ErrorSound();
                }


                if (_stopNetwork == true) break;

            }
        }

       

        #endregion


        #region Визуализация сетевой работы

        // Получение сообщений от клиентов
        public void UpdateReceiveDisplay(int clientnum, string message)
        {
            listBox1.Items.Add("№" + clientnum.ToString() + ": " + message);
        }
        // Делегат доступа к элементу формы listBox1 из вспомогательного потока.
        protected delegate void UpdateReceiveDisplayDelegate(int clientcount, string message);



        public void UpdateClientsDisplay()
        {
            labelCountClient.Text = _countClient.ToString();
        }

        // Делегат доступа к элементу формы labelCountClient из вспомогательного потока.
        protected delegate void UpdateClientsDisplayDelegate();

        // Звуковое оповещение о перехваченной ошибке.
        void ErrorSound()
        {
            Console.Beep(3000, 80);
            Console.Beep(1000, 100);
        }

        #endregion


    }// public partial class TcpServerForm : Form
}
